# -*- coding: utf-8 -*-
#
# Test script for labri instru detect plugin (works whithout Timeside)
#
# Copyright (c) 2014 Dominique Fourer <dominique@fourer.fr>

# This file is part of TimeSide.

# TimeSide is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.

# TimeSide is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TimeSide.  If not, see <http://www.gnu.org/licenses/>.

# Author: D Fourer <dominique@fourer.fr> http://www.fourer.fr


import numpy, scipy #, simpl

from scipy.io import wavfile
import sys 

import my_tools as mt
import os
import timbre_descriptor
import my_lda

# default settings
chemin = 'examples/';
nbits = 16;
MAX_VAL = pow(2.,(nbits-1)) * 1.0;

## cf. python_trainin.m / python_test.m

############### Processing starts here
#n_max 	= 8;                     ## number of principal axis

## Load models
m 			= scipy.io.loadmat('model_inst.mat');
nb_desc		= numpy.squeeze(numpy.array(m['nb_desc']));   ## nb_desc
i_fs 		= numpy.squeeze(numpy.array(m['i_fs']))-1;   ## IRMFSP indices
Vect 		= numpy.squeeze(numpy.array(m['Vect']));   ## Projection basis
n_max		= numpy.shape(Vect)[0];
repr_mu		= numpy.squeeze(numpy.array(m['repr']));
repr_sigma	= numpy.squeeze(numpy.array(m['repr2'])) + 0.0; #+ mt.EPS
inst		= numpy.squeeze(numpy.array(m['inst']));
T_NOISE		= -60;		## noise threshold in dB

#i_fs = numpy.array(i_fs[0:nb_desc]);
#filelist = os.listdir(chemin);

filelist = ["CNRSMH_I_2013_402_001_05_cordo_frotte.wav", "CNRSMH_I_2013_402_001_cordo_pince.wav", "CNRSMH_I_2013_404_001_aero.wav", "CNRSMH_I_2013_405_membrano.wav", "CNRSMH_I_2013_406_001_idiophone_pince.wav"];

#filelist = ["CNRSMH_I_2013_404_001_aero.wav", "CNRSMH_I_2013_405_membrano.wav"];

## debug:
m2 = scipy.io.loadmat('param_ref.mat');
param_ref = numpy.squeeze(numpy.array(m2["param_ref"]))
h_err = numpy.zeros(nb_desc,float);

index = 0;
for f in filelist:
	if f[-3:] == "wav":
		print "Processing", f, "...\n";
		filename = chemin+f;
		## 1 open file and store as a real vector
		Fs, s = wavfile.read(filename);
		s = s/MAX_VAL;

		#if 10 * numpy.log10(mt.rms(s)+timbre_descriptor.EPS) <= T_NOISE:
		#	print " File only contains silence (skipping) \n"
		#	continue;
		
		## 2 compute descriptors
		desc = timbre_descriptor.compute_all_descriptor(s, Fs);
		param_val, field_name = timbre_descriptor.temporalmodeling(desc);
		param_val = param_val[i_fs];
		
		if index == 0:
			field_name_fs = list();
			for i in xrange(0,nb_desc):
				field_name_fs.append(field_name[ i_fs[i]]);
		
		if numpy.isnan(param_val).any() or numpy.isinf(param_val).any():
			print "Contains spurious values", param_val, "\n"
			mt.disp_var(param_val,"param_val");
		
		##debug: force ref values
		#param_val = param_ref[index,:];
		## debug: check values and cummulate error
		h_tmp = timbre_descriptor.check_values(param_val, param_ref[index,i_fs], field_name_fs, Thr=0.2);
		h_err = h_err + h_tmp;
		
		## 3 compare with previous model and select the most probable
		gr1, gr2, p1, p2 = my_lda.pred_lda(numpy.array([param_val, ])+mt.EPS, Vect, repr_mu, repr_sigma);
		i_res = gr1[0];  ## gr1 is euclidean distance criterion
		
		## display result
		famille = inst[i_res][0][0];
		jeu 	= "";
		if i_res > 0:
			jeu = inst[i_res][1][0];
		print "Detected as ", famille, " ", jeu, " according to res1: res1=",gr1[0]," res2=", gr2[0],"p1=",p1[0]," p2=", p2[0],"\n\n";
		index += 1;

print "Top 20 - most discriminant descriptors"
for i in xrange(0,min(20,nb_desc)):
	print i_fs[i],") ", field_name_fs[ i ];
	
print "\n\n\n\n"
	
## debug: sorted list of spurious descriptors
print "Top 20 - spurious descriptors (compared with Matlab computation results)"
a,I = mt.my_sort(h_err, 'descend');
for i in xrange(0,min(20,len(I))):
	print i_fs[i],") ", field_name_fs[ I[i] ], " error:", h_err[I[i]];
